// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use error::Error;
use eval::{self, Num, ResultIndex};
use input::SourceSet;
use node;
use parse::TokenKind;
use rugint::{DivFromAssign, Integer, Pow};
use rugrat::Rational;
use std::cmp;
use std::u32;
use tree::{Index, Tree};

pub fn eval_add(sources: &SourceSet,
                tree: &mut Tree,
                index: Index)
                -> ResultIndex {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_addition(tree, index));
    let lhs_tv = eval::deref_const_tv_check_z(tree, lhs_i)
        .or_else(|_| {
                     gen_err!(tree[lhs_i],
                              sources,
                              "addition operand cannot have z elements")
                 })?;
    let rhs_tv = eval::deref_const_tv_check_z(tree, rhs_i)
        .or_else(|_| {
                     gen_err!(tree[rhs_i],
                              sources,
                              "addition operand cannot have z elements")
                 })?;
    let (lhs_ty_i, lhs_val_i, rhs_ty_i, rhs_val_i) = match (lhs_tv, rhs_tv) {
        (Some((lt, lv)), Some((rt, rv))) => (lt, lv, rt, rv),
        _ => return addition_find_type(sources, tree, index, lhs_i, rhs_i),
    };
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    let result = match (lhs_type, rhs_type) {
        (Num::Int, Num::Int) => add_i_i(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Int, Num::Uns(r)) => {
            add_i_u(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Int, Num::Sig(r)) => {
            add_i_s(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Int, Num::Rat) => add_i_r(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Uns(l), Num::Int) => {
            add_u_i(kind, (l, lhs_val_i), rhs_val_i, tree)
        }
        (Num::Uns(l), Num::Uns(r)) => {
            add_u_u(kind, (l, lhs_val_i), (r, rhs_val_i), tree)
        }
        (Num::Sig(l), Num::Int) => {
            add_s_i(kind, (l, lhs_val_i), rhs_val_i, tree)
        }
        (Num::Sig(l), Num::Sig(r)) => {
            add_s_s(kind, (l, lhs_val_i), (r, rhs_val_i), tree)
        }
        (Num::Rat, Num::Int) => add_r_i(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Rat, Num::Rat) => add_r_r(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Flo, Num::Flo) => add_f_f(kind, lhs_val_i, rhs_val_i, tree),
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot add values of types {} and {}",
                     tree[lhs_ty_i].type_string(),
                     tree[rhs_ty_i].type_string())?
        }
    };
    Ok(Some(eval::replace_with(tree, index, result)))
}

fn addition_find_type(sources: &SourceSet,
                      tree: &mut Tree,
                      index: Index,
                      lhs_i: Index,
                      rhs_i: Index)
                      -> ResultIndex {
    let lhs_ty_i = eval::determine_type(sources, tree, lhs_i)?;
    let rhs_ty_i = eval::determine_type(sources, tree, rhs_i)?;
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    let ty_i = match (lhs_type, rhs_type) {
        (Num::Int, Num::Int) |
        (Num::Int, Num::Rat) |
        (Num::Rat, Num::Rat) |
        (Num::Flo, Num::Flo) => rhs_ty_i,
        (Num::Rat, Num::Int) => lhs_ty_i,
        (Num::Int, Num::Uns(_)) |
        (Num::Int, Num::Sig(_)) => {
            eval::cast_type(sources, tree, lhs_i, rhs_ty_i)?;
            rhs_ty_i
        }
        (Num::Uns(_), Num::Int) |
        (Num::Sig(_), Num::Int) => {
            eval::cast_type(sources, tree, rhs_i, lhs_ty_i)?;
            lhs_ty_i
        }
        (Num::Uns(l), Num::Uns(r)) |
        (Num::Sig(l), Num::Sig(r)) => {
            eval::cast_to_bigger(sources,
                                 tree,
                                 (l, lhs_i, lhs_ty_i),
                                 (r, rhs_i, rhs_ty_i))?
        }
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot add values of types {} and {}",
                     tree[lhs_ty_i].type_string(),
                     tree[rhs_ty_i].type_string())?
        }
    };
    Ok(Some(eval::set_type(tree, index, ty_i)))
}

fn add_int(op: TokenKind,
           lhs_i: Index,
           lhs_signed_size: Option<u32>,
           rhs_i: Index,
           rhs_signed_size: Option<u32>,
           tree: &Tree)
           -> Integer {
    let (lhs_val, _) = node::get_integer_val(tree, lhs_i).unwrap();
    let (rhs_val, _) = node::get_integer_val(tree, rhs_i).unwrap();
    let a = if let Some(size) = lhs_signed_size {
        eval::sign_extend(lhs_val, size)
    } else {
        lhs_val.clone()
    };
    let neg_rhs = rhs_signed_size.and_then(|size| {
                                               eval::sign_extend_if_neg(rhs_val,
                                                                        size)
                                           });
    let b = match neg_rhs {
        Some(ref r) => r,
        None => rhs_val,
    };
    match op {
        TokenKind::Plus => a + b,
        TokenKind::Minus => a - b,
        _ => unreachable!(),
    }
}

fn add_i_i(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = add_int(op, lhs_i, None, rhs_i, None, tree);
    node::typed_value_integer(val, 0.into(), tree)
}

fn add_i_u(op: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = add_int(op, lhs_i, None, rhs_i, None, tree);
    node::typed_value_unsigned(rhs_size, val, 0.into(), tree)
}

fn add_u_i(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = add_int(op, lhs_i, None, rhs_i, None, tree);
    node::typed_value_unsigned(lhs_size, val, 0.into(), tree)
}

fn add_u_u(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = add_int(op, lhs_i, None, rhs_i, None, tree);
    node::typed_value_unsigned(cmp::max(lhs_size, rhs_size),
                               val,
                               0.into(),
                               tree)
}

fn add_i_s(op: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = add_int(op, lhs_i, None, rhs_i, Some(rhs_size), tree);
    node::typed_value_unsigned(rhs_size, val, 0.into(), tree)
}

fn add_s_i(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = add_int(op, lhs_i, Some(lhs_size), rhs_i, None, tree);
    node::typed_value_signed(lhs_size, val, 0.into(), tree)
}

fn add_s_s(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = add_int(op, lhs_i, Some(lhs_size), rhs_i, Some(rhs_size), tree);
    node::typed_value_signed(cmp::max(lhs_size, rhs_size), val, 0.into(), tree)
}

fn add_i_r(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let (lhs_val, _) = node::get_integer_val(tree, lhs_i).unwrap();
        let rhs_val = node::get_rational_val(tree, rhs_i).unwrap();
        match op {
            TokenKind::Plus => Rational::from(lhs_val) + rhs_val,
            TokenKind::Minus => Rational::from(lhs_val) - rhs_val,
            _ => unreachable!(),
        }
    };
    eval::rational_or_integer(val, tree)
}

fn add_r_i(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let lhs_val = node::get_rational_val(tree, lhs_i).unwrap();
        let (rhs_val, _) = node::get_integer_val(tree, rhs_i).unwrap();
        match op {
            TokenKind::Plus => Rational::from(rhs_val) + lhs_val,
            TokenKind::Minus => -Rational::from(rhs_val) + lhs_val,
            _ => unreachable!(),
        }
    };
    eval::rational_or_integer(val, tree)
}

fn add_r_r(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let lhs_val = node::get_rational_val(tree, lhs_i).unwrap();
        let rhs_val = node::get_rational_val(tree, rhs_i).unwrap();
        match op {
            TokenKind::Plus => lhs_val.clone() + rhs_val,
            TokenKind::Minus => lhs_val.clone() - rhs_val,
            _ => unreachable!(),
        }
    };
    eval::rational_or_integer(val, tree)
}

fn add_f_f(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let lhs_val = node::get_float_val(tree, lhs_i).unwrap();
        let rhs_val = node::get_float_val(tree, rhs_i).unwrap();
        match op {
            TokenKind::Plus => lhs_val.clone() + rhs_val,
            TokenKind::Minus => lhs_val.clone() - rhs_val,
            _ => unreachable!(),
        }
    };
    node::typed_value_float(val, tree)
}

pub fn eval_mul(sources: &SourceSet,
                tree: &mut Tree,
                index: Index)
                -> ResultIndex {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_multiplication(tree, index));
    let lhs_tv = eval::deref_const_tv_check_z(tree, lhs_i)
        .or_else(|_| {
                     gen_err!(tree[lhs_i],
                              sources,
                              "multiplication/division operand cannot have z \
                          elements")
                 })?;
    let rhs_tv = eval::deref_const_tv_check_z(tree, rhs_i)
        .or_else(|_| {
                     gen_err!(tree[rhs_i],
                              sources,
                              "multiplication/division operand cannot have z \
                          elements")
                 })?;
    if (kind == TokenKind::Division || kind == TokenKind::Remainder) &&
       eval::is_nonfloat_zero(tree, rhs_tv) {
        gen_err!(tree[rhs_i], sources, "division by zero")?
    }
    if kind == TokenKind::Division && eval::is_float_zero(tree, lhs_tv) &&
       eval::is_float_zero(tree, rhs_tv) {
        gen_err!(tree[rhs_i], sources, "zero divided by zero")?
    }
    let (lhs_ty_i, lhs_val_i, rhs_ty_i, rhs_val_i) = match (lhs_tv, rhs_tv) {
        (Some((lt, lv)), Some((rt, rv))) => (lt, lv, rt, rv),
        _ => {
            return multiplication_find_type(sources,
                                            tree,
                                            kind,
                                            index,
                                            lhs_i,
                                            rhs_i)
        }
    };
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    if kind == TokenKind::Remainder &&
       (!lhs_type.is_integral() || !rhs_type.is_integral()) {
        gen_err!(tree[index],
                 sources,
                 "remainder takes integer operands only")?
    }
    let result = match (lhs_type, rhs_type) {
        (Num::Int, Num::Int) => mul_i_i(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Int, Num::Uns(r)) => {
            mul_i_u(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Int, Num::Sig(r)) => {
            mul_i_s(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Int, Num::Rat) => mul_i_r(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Uns(l), Num::Int) => {
            mul_u_i(kind, (l, lhs_val_i), rhs_val_i, tree)
        }
        (Num::Uns(l), Num::Uns(r)) => {
            mul_u_u(kind, (l, lhs_val_i), (r, rhs_val_i), tree)
        }
        (Num::Sig(l), Num::Int) => {
            mul_s_i(kind, (l, lhs_val_i), rhs_val_i, tree)
        }
        (Num::Sig(l), Num::Uns(r)) => {
            mul_s_s(kind, (l, lhs_val_i), (r, rhs_val_i), tree)
        }
        (Num::Rat, Num::Int) => mul_r_i(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Rat, Num::Rat) => mul_r_r(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Flo, Num::Flo) => mul_f_f(kind, lhs_val_i, rhs_val_i, tree),
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot multiply/divide values of types {} and {}",
                     tree[lhs_ty_i].type_string(),
                     tree[rhs_ty_i].type_string())?
        }
    };
    Ok(Some(eval::replace_with(tree, index, result)))
}

fn multiplication_find_type(sources: &SourceSet,
                            tree: &mut Tree,
                            kind: TokenKind,
                            index: Index,
                            lhs_i: Index,
                            rhs_i: Index)
                            -> ResultIndex {
    let mut lhs_ty_i = eval::determine_type(sources, tree, lhs_i)?;
    let mut rhs_ty_i = eval::determine_type(sources, tree, rhs_i)?;
    let mut lhs_type = eval::num_type(tree, lhs_ty_i);
    let mut rhs_type = eval::num_type(tree, rhs_ty_i);
    if kind == TokenKind::Remainder &&
       (!lhs_type.is_integral() || !rhs_type.is_integral()) {
        gen_err!(tree[index],
                 sources,
                 "remainder takes integer operands only")?
    }
    match (lhs_type, rhs_type) {
        (Num::Int, Num::Uns(_)) |
        (Num::Int, Num::Sig(_)) => {
            let cast_ty_i = eval::cast_type(sources, tree, lhs_i, rhs_ty_i)
                .unwrap();
            lhs_ty_i = cast_ty_i;
            lhs_type = eval::num_type(tree, lhs_ty_i);
        }
        (Num::Uns(_), Num::Int) |
        (Num::Sig(_), Num::Int) => {
            let cast_ty_i = eval::cast_type(sources, tree, rhs_i, lhs_ty_i)
                .unwrap();
            rhs_ty_i = cast_ty_i;
            rhs_type = eval::num_type(tree, rhs_ty_i);
        }
        _ => {}
    }
    let ty_i = match (lhs_type, rhs_type, kind) {
        (Num::Uns(l), Num::Uns(r), TokenKind::Times) => {
            node::unsigned_type(l + r, tree)
        }
        (Num::Sig(l), Num::Sig(r), TokenKind::Times) => {
            node::signed_type(l + r, tree)
        }
        (Num::Int, Num::Int, _) |
        (Num::Rat, Num::Int, _) |
        (Num::Rat, Num::Rat, _) |
        (Num::Flo, Num::Flo, _) |
        (Num::Uns(_), Num::Uns(_), TokenKind::Division) |
        (Num::Sig(_), Num::Sig(_), TokenKind::Division) => lhs_ty_i,
        (Num::Int, Num::Rat, _) |
        (Num::Uns(_), Num::Uns(_), TokenKind::Remainder) |
        (Num::Sig(_), Num::Sig(_), TokenKind::Remainder) => rhs_ty_i,
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot multiply/divide values of types {} and {}",
                     tree[lhs_ty_i].type_string(),
                     tree[rhs_ty_i].type_string())?
        }
    };
    Ok(Some(eval::set_type(tree, index, ty_i)))
}

fn mul_int(op: TokenKind,
           lhs_i: Index,
           lhs_signed_size: Option<u32>,
           rhs_i: Index,
           rhs_signed_size: Option<u32>,
           tree: &Tree)
           -> Integer {
    let (lhs_val, _) = node::get_integer_val(tree, lhs_i).unwrap();
    let (rhs_val, _) = node::get_integer_val(tree, rhs_i).unwrap();
    let a = if let Some(size) = lhs_signed_size {
        eval::sign_extend(lhs_val, size)
    } else {
        lhs_val.clone()
    };
    let neg_rhs = rhs_signed_size.and_then(|size| {
                                               eval::sign_extend_if_neg(rhs_val,
                                                                        size)
                                           });
    let b = match neg_rhs {
        Some(ref r) => r,
        None => rhs_val,
    };
    match op {
        TokenKind::Times => a * b,
        TokenKind::Division => a / b,
        TokenKind::Remainder => a % b,
        _ => unreachable!(),
    }
}

fn mul_i_i(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    if op == TokenKind::Division {
        let (num, den) = {
            let (num, _) = node::get_integer_val(tree, lhs_i).unwrap();
            let (den, _) = node::get_integer_val(tree, rhs_i).unwrap();
            (num.clone(), den.clone())
        };
        eval::rational_or_integer(Rational::from((num, den)), tree)
    } else {
        let val = mul_int(op, lhs_i, None, rhs_i, None, tree);
        node::typed_value_integer(val, 0.into(), tree)
    }
}

fn mul_i_u(op: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = mul_int(op, lhs_i, None, rhs_i, None, tree);
    let size = match op {
        TokenKind::Times => rhs_size * 2,
        _ => rhs_size,
    };
    node::typed_value_unsigned(size, val, 0.into(), tree)
}

fn mul_u_i(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = mul_int(op, lhs_i, None, rhs_i, None, tree);
    let size = match op {
        TokenKind::Times => lhs_size * 2,
        _ => lhs_size,
    };
    node::typed_value_unsigned(size, val, 0.into(), tree)
}

fn mul_i_s(op: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = mul_int(op, lhs_i, None, rhs_i, Some(rhs_size), tree);
    let size = match op {
        TokenKind::Times => rhs_size * 2,
        _ => rhs_size,
    };
    node::typed_value_signed(size, val, 0.into(), tree)
}

fn mul_s_i(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = mul_int(op, lhs_i, Some(lhs_size), rhs_i, None, tree);
    let size = match op {
        TokenKind::Times => lhs_size * 2,
        _ => lhs_size,
    };
    node::typed_value_signed(size, val, 0.into(), tree)
}

fn mul_i_r(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let (lhs_val, _) = node::get_integer_val(tree, lhs_i).unwrap();
        let rhs_val = node::get_rational_val(tree, rhs_i).unwrap();
        match op {
            TokenKind::Times => Rational::from(lhs_val) * rhs_val,
            TokenKind::Division => Rational::from(lhs_val) / rhs_val,
            _ => unreachable!(),
        }
    };
    eval::rational_or_integer(val, tree)
}

fn mul_r_i(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let lhs_val = node::get_rational_val(tree, lhs_i).unwrap();
        let (rhs_val, _) = node::get_integer_val(tree, rhs_i).unwrap();
        let mut ans = Rational::from(rhs_val);
        match op {
            TokenKind::Times => ans *= lhs_val,
            TokenKind::Division => ans.div_from_assign(lhs_val),
            _ => unreachable!(),
        };
        ans
    };
    eval::rational_or_integer(val, tree)
}

fn mul_u_u(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = mul_int(op, lhs_i, None, rhs_i, None, tree);
    let size = match op {
        TokenKind::Times => lhs_size + rhs_size,
        TokenKind::Division => lhs_size,
        TokenKind::Remainder => rhs_size,
        _ => unreachable!(),
    };
    node::typed_value_unsigned(size, val, 0.into(), tree)
}

fn mul_s_s(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = mul_int(op, lhs_i, Some(lhs_size), rhs_i, Some(rhs_size), tree);
    let size = match op {
        TokenKind::Times => lhs_size + rhs_size,
        TokenKind::Division => lhs_size,
        TokenKind::Remainder => rhs_size,
        _ => unreachable!(),
    };
    node::typed_value_signed(size, val, 0.into(), tree)
}

fn mul_r_r(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let lhs_val = node::get_rational_val(tree, lhs_i).unwrap();
        let rhs_val = node::get_rational_val(tree, rhs_i).unwrap();
        match op {
            TokenKind::Times => lhs_val.clone() * rhs_val,
            TokenKind::Division => lhs_val.clone() / rhs_val,
            _ => unreachable!(),
        }
    };
    eval::rational_or_integer(val, tree)
}

fn mul_f_f(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = {
        let lhs_val = node::get_float_val(tree, lhs_i).unwrap();
        let rhs_val = node::get_float_val(tree, rhs_i).unwrap();
        match op {
            TokenKind::Times => lhs_val.clone() * rhs_val,
            TokenKind::Division => lhs_val.clone() / rhs_val,
            _ => unreachable!(),
        }
    };
    node::typed_value_float(val, tree)
}

pub fn eval_pow(sources: &SourceSet,
                tree: &mut Tree,
                index: Index)
                -> ResultIndex {
    let (lhs_i, rhs_i) = unwrap_or_try_next!(node::get_expr_power(tree, index));
    let lhs_tv = eval::deref_const_tv_check_z(tree, lhs_i)
        .or_else(|_| {
                     gen_err!(tree[lhs_i],
                              sources,
                              "shift operand cannot have z elements")
                 })?;
    let rhs_tv = eval::deref_const_tv_check_z(tree, rhs_i)
        .or_else(|_| {
                     gen_err!(tree[rhs_i],
                              sources,
                              "shift operand cannot have z elements")
                 })?;
    let (lhs_ty_i, lhs_val_i, rhs_ty_i, rhs_val_i) = match (lhs_tv, rhs_tv) {
        (Some((lt, lv)), Some((rt, rv))) => (lt, lv, rt, rv),
        _ => return power_find_type(sources, tree, index, lhs_i, rhs_i),
    };
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    let int_power = if rhs_type.is_integral() {
        let (rhs_int, _) = node::get_integer_val(tree, rhs_val_i).unwrap();
        let (sign, abs) = match eval::get_sign_abs(rhs_int,
                                                   rhs_type,
                                                   node::MAX_UNSIGNED_SIZE) {
            Some(sa) => sa,
            None => {
                gen_err!(tree[rhs_val_i],
                         sources,
                         "power operand is too large")?
            }
        };
        if sign && lhs_type.is_integral() && lhs_type != Num::Int {
            gen_err!(tree[rhs_i],
                     sources,
                     "power operand cannot be negative for sized integers")?;
        }
        if sign && eval::is_nonfloat_zero(tree, lhs_tv) {
            gen_err!(tree[index], sources, "negative power of zero")?;
        }
        let signed = if sign { -(abs as i32) } else { abs as i32 };
        Some((abs, signed))
    } else {
        None
    };
    let result = match (lhs_type, int_power) {
        (Num::Int, Some((_, signed))) => power_i(lhs_val_i, signed, tree),
        (Num::Uns(size), Some((abs, _))) => {
            power_u((size, lhs_val_i), abs, tree)
        }
        (Num::Sig(size), Some((abs, _))) => {
            power_s((size, lhs_val_i), abs, tree)
        }
        (Num::Rat, Some((_, signed))) => power_r(lhs_val_i, signed, tree),
        (Num::Flo, Some((_, signed))) => power_f_i(lhs_val_i, signed, tree),
        (Num::Flo, _) if rhs_type == Num::Flo => {
            power_f_f(sources, index, lhs_val_i, rhs_val_i, tree)?
        }
        _ => {
            gen_err!(tree[index],
                     sources,
                     "unsupported operand types to power operator")?
        }
    };
    Ok(Some(eval::replace_with(tree, index, result)))
}

fn power_find_type(sources: &SourceSet,
                   tree: &mut Tree,
                   index: Index,
                   lhs_i: Index,
                   rhs_i: Index)
                   -> ResultIndex {
    let lhs_ty_i = eval::determine_type(sources, tree, lhs_i)?;
    let rhs_ty_i = eval::determine_type(sources, tree, rhs_i)?;
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    let rhs_integral = rhs_type.is_integral();
    let ty_i = match lhs_type {
        Num::Int if rhs_integral => node::rational_type(tree),
        Num::Uns(_) | Num::Sig(_) | Num::Rat | Num::Flo if rhs_integral => {
            lhs_ty_i
        }
        Num::Flo if rhs_type == Num::Flo => lhs_ty_i,
        _ => {
            gen_err!(tree[index],
                     sources,
                     "unsupported operand types to power operator")?
        }
    };
    Ok(Some(eval::set_type(tree, index, ty_i)))
}

fn power_i(lhs_i: Index, rhs: i32, tree: &mut Tree) -> Index {
    let lhs = {
        let (lhs, _) = node::get_integer_val(tree, lhs_i).unwrap();
        Rational::from(lhs)
    };
    eval::rational_or_integer(lhs.pow(rhs), tree)
}

fn power_u((lhs_size, lhs_i): (u32, Index),
           rhs: u32,
           tree: &mut Tree)
           -> Index {
    let lhs = {
        let (lhs, _) = node::get_integer_val(tree, lhs_i).unwrap();
        lhs.clone()
    };
    node::typed_value_unsigned(lhs_size, lhs.pow(rhs), 0.into(), tree)
}

fn power_s((lhs_size, lhs_i): (u32, Index),
           rhs: u32,
           tree: &mut Tree)
           -> Index {
    let lhs = {
        let (lhs, _) = node::get_integer_val(tree, lhs_i).unwrap();
        eval::sign_extend(lhs, lhs_size)
    };
    node::typed_value_signed(lhs_size, lhs.pow(rhs), 0.into(), tree)
}

fn power_r(lhs_i: Index, rhs: i32, tree: &mut Tree) -> Index {
    let lhs = node::get_rational_val(tree, lhs_i).unwrap().clone();
    eval::rational_or_integer(lhs.pow(rhs), tree)
}

fn power_f_f(sources: &SourceSet,
             index: Index,
             lhs_i: Index,
             rhs_i: Index,
             tree: &mut Tree)
             -> Result<Index, Error> {
    let lhs = node::get_float_val(tree, lhs_i).unwrap().clone();
    let val = lhs.pow(node::get_float_val(tree, rhs_i).unwrap());
    if val.is_nan() {
        gen_err!(tree[index], sources, "float operation produces NaN")?;
    }
    Ok(node::typed_value_float(val, tree))
}

fn power_f_i(lhs_i: Index, rhs: i32, tree: &mut Tree) -> Index {
    let lhs = node::get_float_val(tree, lhs_i).unwrap().clone();
    node::typed_value_float(lhs.pow(rhs), tree)
}

pub fn eval_unary(sources: &SourceSet,
                  tree: &mut Tree,
                  index: Index)
                  -> ResultIndex {
    let (kind, op_i) = unwrap_or_try_next!(node::get_expr_unary(tree, index));
    let op_tv = eval::deref_const_tv_check_z(tree, op_i)
        .or_else(|_| {
                     gen_err!(tree[op_i],
                              sources,
                              "unary operand cannot have z elements")
                 })?;
    let (op_ty_i, op_val_i) = match op_tv {
        Some((t, v)) => (t, v),
        _ => return unary_find_type(sources, tree, index, kind, op_i),
    };
    let op_type = eval::num_type(tree, op_ty_i);
    if kind != TokenKind::Plus && kind != TokenKind::Minus &&
       !op_type.is_integral() {
        gen_err!(tree[op_ty_i],
                 sources,
                 "unary bitwise operand must be of an integer type, not {}",
                 tree[op_ty_i].type_string())?
    }
    if kind == TokenKind::Xor && op_type == Num::Int {
        let (val, _) = node::get_integer_val(tree, op_val_i).unwrap();
        if *val < 0 {
            gen_err!(tree[op_val_i],
                     sources,
                     "cannot perform reduction xor on unsized negative \
                      integer")?;
        }
    }
    let result = match op_type {
        Num::Int => unary_i(kind, op_val_i, tree),
        Num::Uns(size) => unary_u(kind, size, op_val_i, tree),
        Num::Sig(size) => unary_s(kind, size, op_val_i, tree),
        Num::Rat => unary_r(kind, op_val_i, tree),
        Num::Flo => unary_f(kind, op_val_i, tree),
        _ => {
            gen_err!(tree[op_ty_i],
                     sources,
                     "unary operand must be of a numeric type, not {}",
                     tree[op_ty_i].type_string())?
        }
    };
    Ok(Some(eval::replace_with(tree, index, result)))
}

fn unary_find_type(sources: &SourceSet,
                   tree: &mut Tree,
                   index: Index,
                   kind: TokenKind,
                   op_i: Index)
                   -> ResultIndex {
    let op_ty_i = eval::determine_type(sources, tree, op_i)?;
    let op_type = eval::num_type(tree, op_ty_i);
    if op_type == Num::Not {
        gen_err!(tree[op_ty_i],
                 sources,
                 "unary operand must be of a numeric type, not {}",
                 tree[op_ty_i].type_string())?;
    }
    if kind != TokenKind::Plus && kind != TokenKind::Minus &&
       !op_type.is_integral() {
        gen_err!(tree[op_ty_i],
                 sources,
                 "unary bitwise operand must be of an integer type, not {}",
                 tree[op_ty_i].type_string())?
    }
    let ty_i = match kind {
        TokenKind::Plus | TokenKind::Minus | TokenKind::Not => op_ty_i,
        _ if op_type == Num::Int => {
            gen_err!(tree[op_ty_i],
                     sources,
                     "reduction operation not supported on unsized \
                      non-constant integer")?
        }
        _ => node::bit_type(tree),
    };
    let ret = eval::set_type(tree, index, ty_i);
    Ok(Some(ret))
}

fn unary_non_reduction(op: TokenKind,
                       index: Index,
                       tree: &mut Tree)
                       -> Integer {
    let (val, _) = node::get_integer_val(tree, index).unwrap();
    match op {
        TokenKind::Plus => val.clone(),
        TokenKind::Minus => -val.clone(),
        TokenKind::Not => !val.clone(),
        _ => unreachable!(),
    }
}

fn unary_reduction(op: TokenKind, index: Index, tree: &mut Tree) -> Index {
    let bool_val = {
        let (val, _) = node::get_integer_val(tree, index).unwrap();
        match op {
            TokenKind::And => *val == -1,
            TokenKind::Or => *val != 0,
            TokenKind::Xor => (val.count_ones().unwrap() & 1) != 0,
            _ => unreachable!(),
        }
    };
    node::typed_value_bit(bool_val, tree)
}

fn unary_i(op: TokenKind, index: Index, tree: &mut Tree) -> Index {
    match op {
        TokenKind::Plus | TokenKind::Minus | TokenKind::Not => {
            let val = unary_non_reduction(op, index, tree);
            node::typed_value_integer(val, 0.into(), tree)
        }
        _ => unary_reduction(op, index, tree),
    }
}

fn unary_u(op: TokenKind, size: u32, index: Index, tree: &mut Tree) -> Index {
    match op {
        TokenKind::Plus | TokenKind::Minus | TokenKind::Not => {
            let val = unary_non_reduction(op, index, tree);
            node::typed_value_unsigned(size, val, 0.into(), tree)
        }
        _ => unary_reduction(op, index, tree),
    }
}

fn unary_s(op: TokenKind, size: u32, index: Index, tree: &mut Tree) -> Index {
    match op {
        TokenKind::Plus | TokenKind::Minus | TokenKind::Not => {
            let val = unary_non_reduction(op, index, tree);
            node::typed_value_signed(size, val, 0.into(), tree)
        }
        _ => unary_reduction(op, index, tree),
    }
}

fn unary_r(op: TokenKind, index: Index, tree: &mut Tree) -> Index {
    let ans = {
        let val = node::get_rational_val(tree, index).unwrap();
        match op {
            TokenKind::Plus => val.clone(),
            TokenKind::Minus => -val.clone(),
            _ => unreachable!(),
        }
    };
    node::typed_value_rational(ans, tree)
}

fn unary_f(op: TokenKind, index: Index, tree: &mut Tree) -> Index {
    let ans = {
        let val = node::get_float_val(tree, index).unwrap();
        match op {
            TokenKind::Plus => val.clone(),
            TokenKind::Minus => -val.clone(),
            _ => unreachable!(),
        }
    };
    node::typed_value_float(ans, tree)
}
