// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use error::Error;
use eval::{self, Num, ResultIndex};
use input::SourceSet;
use node;
use parse::TokenKind;
use rugint::Integer;
use std::cmp;
use std::u32;
use tree::{Index, Tree};

pub fn eval(sources: &SourceSet, tree: &mut Tree, index: Index) -> ResultIndex {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_bitwise(tree, index));
    let lhs_tv = eval::deref_const_tv_check_z(tree, lhs_i)
        .or_else(|_| {
                     gen_err!(tree[lhs_i],
                              sources,
                              "bitwise operand cannot have z elements")
                 })?;
    let rhs_tv = eval::deref_const_tv_check_z(tree, rhs_i)
        .or_else(|_| {
                     gen_err!(tree[rhs_i],
                              sources,
                              "bitwise operand cannot have z elements")
                 })?;
    let (lhs_ty_i, lhs_val_i, rhs_ty_i, rhs_val_i) = match (lhs_tv, rhs_tv) {
        (Some((lt, lv)), Some((rt, rv))) => (lt, lv, rt, rv),
        _ => return bitwise_find_type(sources, tree, index, lhs_i, rhs_i),
    };
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    if !lhs_type.is_integral() {
        gen_err!(tree[lhs_i],
                 sources,
                 "bitwise operand must be of an integer type, not {}",
                 tree[lhs_ty_i].type_string())?;
    };
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    if !rhs_type.is_integral() {
        gen_err!(tree[rhs_i],
                 sources,
                 "bitwise operand must be of an integer type, not {}",
                 tree[rhs_ty_i].type_string())?;
    };
    let result = match (lhs_type, rhs_type) {
        (Num::Int, Num::Int) => bit_i_i(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Int, Num::Uns(r)) => {
            bit_i_u(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Int, Num::Sig(r)) => {
            bit_i_s(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Uns(l), Num::Int) => {
            bit_i_u(kind, rhs_val_i, (l, lhs_val_i), tree)
        }
        (Num::Uns(l), Num::Uns(r)) => {
            bit_u_u(kind, (l, lhs_val_i), (r, rhs_val_i), tree)
        }
        (Num::Sig(l), Num::Int) => {
            bit_i_s(kind, rhs_val_i, (l, lhs_val_i), tree)
        }
        (Num::Sig(l), Num::Sig(r)) => {
            bit_s_s(kind, (l, lhs_val_i), (r, rhs_val_i), tree)
        }
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot mix unsigned and signed operands in bitwise \
                      operation")?
        }
    };
    Ok(Some(eval::replace_with(tree, index, result)))
}

fn bitwise_find_type(sources: &SourceSet,
                     tree: &mut Tree,
                     index: Index,
                     lhs_i: Index,
                     rhs_i: Index)
                     -> ResultIndex {
    let lhs_ty_i = eval::determine_type(sources, tree, lhs_i)?;
    let rhs_ty_i = eval::determine_type(sources, tree, rhs_i)?;
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    if !lhs_type.is_integral() {
        gen_err!(tree[lhs_i],
                 sources,
                 "bitwise operand must be of an integer type, not {}",
                 tree[lhs_ty_i].type_string())?;
    };
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    if !rhs_type.is_integral() {
        gen_err!(tree[rhs_i],
                 sources,
                 "bitwise operand must be of an integer type, not {}",
                 tree[rhs_ty_i].type_string())?;
    };
    let ty_i = match (lhs_type, rhs_type) {
        (Num::Int, Num::Int) => lhs_ty_i,
        (Num::Int, _) => {
            eval::cast_type(sources, tree, lhs_i, rhs_ty_i)?;
            rhs_ty_i
        }
        (_, Num::Int) => {
            eval::cast_type(sources, tree, rhs_i, lhs_ty_i)?;
            lhs_ty_i
        }
        (Num::Uns(l), Num::Uns(r)) |
        (Num::Sig(l), Num::Sig(r)) => {
            eval::cast_to_bigger(sources,
                                 tree,
                                 (l, lhs_i, lhs_ty_i),
                                 (r, rhs_i, rhs_ty_i))?
        }
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot mix unsigned and signed operands in bitwise \
                      operation")?
        }
    };
    Ok(Some(eval::set_type(tree, index, ty_i)))
}

fn bit_int(op: TokenKind,
           lhs_i: Index,
           lhs_signed_size: Option<u32>,
           rhs_i: Index,
           rhs_signed_size: Option<u32>,
           tree: &Tree)
           -> Integer {
    let (lhs_val, _) = node::get_integer_val(tree, lhs_i).unwrap();
    let (rhs_val, _) = node::get_integer_val(tree, rhs_i).unwrap();
    let neg_lhs = lhs_signed_size.and_then(|size| {
                                               eval::sign_extend_if_neg(lhs_val,
                                                                        size)
                                           });
    let neg_rhs = rhs_signed_size.and_then(|size| {
                                               eval::sign_extend_if_neg(rhs_val,
                                                                        size)
                                           });
    let (a, b) = if let Some(l) = neg_lhs {
        match neg_rhs {
            Some(ref r) => (l, r),
            None => (l, rhs_val),
        }
    } else {
        match neg_rhs {
            Some(r) => (r, lhs_val),
            None => (lhs_val.clone(), rhs_val),
        }
    };
    match op {
        TokenKind::And => a & b,
        TokenKind::Or => a | b,
        TokenKind::Xor => a ^ b,
        TokenKind::Nand => !(a & b),
        TokenKind::Nor => !(a | b),
        TokenKind::Xnor => !(a ^ b),
        _ => unreachable!(),
    }
}

fn bit_i_i(op: TokenKind,
           lhs_i: Index,
           rhs_i: Index,
           tree: &mut Tree)
           -> Index {
    let val = bit_int(op, lhs_i, None, rhs_i, None, tree);
    node::typed_value_integer(val, 0.into(), tree)
}

fn bit_i_u(op: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = bit_int(op, lhs_i, None, rhs_i, None, tree);
    node::typed_value_unsigned(rhs_size, val, 0.into(), tree)
}

fn bit_i_s(op: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = bit_int(op, lhs_i, None, rhs_i, Some(rhs_size), tree);
    node::typed_value_signed(rhs_size, val, 0.into(), tree)
}

fn bit_u_u(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = bit_int(op, lhs_i, None, rhs_i, None, tree);
    let size = cmp::max(lhs_size, rhs_size);
    node::typed_value_unsigned(size, val, 0.into(), tree)
}

fn bit_s_s(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           (rhs_size, rhs_i): (u32, Index),
           tree: &mut Tree)
           -> Index {
    let val = bit_int(op, lhs_i, Some(lhs_size), rhs_i, Some(rhs_size), tree);
    let size = cmp::max(lhs_size, rhs_size);
    node::typed_value_signed(size, val, 0.into(), tree)
}

pub fn eval_cat(sources: &SourceSet,
                tree: &mut Tree,
                index: Index)
                -> ResultIndex {
    let (lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_concatenation(tree, index));
    let lhs_tv = eval::deref_const_tv(tree, lhs_i);
    let rhs_tv = eval::deref_const_tv(tree, rhs_i);
    let (lhs_ty_i, lhs_val_i, rhs_ty_i, rhs_val_i) = match (lhs_tv, rhs_tv) {
        (Some((lt, lv)), Some((rt, rv))) => (lt, lv, rt, rv),
        _ => return concatenation_find_type(sources, tree, index, lhs_i, rhs_i),
    };
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let lhs_size = match lhs_type {
        Num::Uns(size) | Num::Sig(size) => size,
        _ => {
            gen_err!(tree[lhs_i],
                     sources,
                     "concatenation operand must be of a sized integer type, \
                      not {}",
                     tree[lhs_ty_i].type_string())?
        }
    };
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    let rhs_size = match (lhs_type, rhs_type) {
        (Num::Uns(_), Num::Uns(size)) |
        (Num::Sig(_), Num::Sig(size)) => size,
        (_, Num::Sig(_)) | (_, Num::Uns(_)) => {
            gen_err!(tree[index],
                     sources,
                     "cannot mix unsigned and signed operands in \
                      concatenation operation")?
        }
        _ => {
            gen_err!(tree[rhs_i],
                     sources,
                     "concatenation operand must be of a sized integer type, \
                      not {}",
                     tree[rhs_ty_i].type_string())?
        }
    };
    if lhs_size + rhs_size > node::MAX_UNSIGNED_SIZE {
        gen_err!(tree[index],
                 sources,
                 "concatentated expression is too large")?;
    }
    let result = match lhs_type {
        Num::Uns(_) => {
            cat_u((lhs_size, lhs_val_i), (rhs_size, rhs_val_i), tree)
        }
        Num::Sig(_) => {
            cat_s((lhs_size, lhs_val_i), (rhs_size, rhs_val_i), tree)
        }
        _ => unreachable!(),
    };
    Ok(Some(eval::replace_with(tree, index, result)))
}

fn concatenation_find_type(sources: &SourceSet,
                           tree: &mut Tree,
                           index: Index,
                           lhs_i: Index,
                           rhs_i: Index)
                           -> ResultIndex {
    let lhs_ty_i = eval::determine_type(sources, tree, lhs_i)?;
    let rhs_ty_i = eval::determine_type(sources, tree, rhs_i)?;
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let lhs_size = match lhs_type {
        Num::Uns(size) | Num::Sig(size) => size,
        _ => {
            gen_err!(tree[lhs_i],
                     sources,
                     "concatenation operand must be of a sized integer type, \
                      not {}",
                     tree[lhs_ty_i].type_string())?
        }
    };
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    let rhs_size = match (lhs_type, rhs_type) {
        (Num::Uns(_), Num::Uns(size)) |
        (Num::Sig(_), Num::Sig(size)) => size,
        (_, Num::Sig(_)) | (_, Num::Uns(_)) => {
            gen_err!(tree[index],
                     sources,
                     "cannot mix unsigned and signed operands in \
                      concatenation operation")?
        }
        _ => {
            gen_err!(tree[rhs_i],
                     sources,
                     "concatenation operand must be of a sized integer type, \
                      not {}",
                     tree[rhs_ty_i].type_string())?
        }
    };
    if lhs_size + rhs_size > node::MAX_UNSIGNED_SIZE {
        gen_err!(tree[index],
                 sources,
                 "concatentated expression is too large")?;
    }
    let ty_i = match lhs_type {
        Num::Uns(_) => node::unsigned_type(lhs_size + rhs_size, tree),
        Num::Sig(_) => node::signed_type(lhs_size + rhs_size, tree),
        _ => unreachable!(),
    };
    Ok(Some(eval::set_type(tree, index, ty_i)))

}

fn cat(lhs_i: Index,
       (rhs_size, rhs_i): (u32, Index),
       tree: &Tree)
       -> (Integer, Integer) {
    let (lhs_v, lhs_m) = node::get_integer_val(tree, lhs_i).unwrap();
    let (rhs_v, rhs_m) = node::get_integer_val(tree, rhs_i).unwrap();
    ((lhs_v.clone() << rhs_size) | rhs_v, (lhs_m.clone() << rhs_size) | rhs_m)
}

fn cat_u((lhs_size, lhs_i): (u32, Index),
         (rhs_size, rhs_i): (u32, Index),
         tree: &mut Tree)
         -> Index {
    let (val, mode) = cat(lhs_i, (rhs_size, rhs_i), tree);
    node::typed_value_unsigned(lhs_size + rhs_size, val, mode, tree)
}

fn cat_s((lhs_size, lhs_i): (u32, Index),
         (rhs_size, rhs_i): (u32, Index),
         tree: &mut Tree)
         -> Index {
    let (val, mode) = cat(lhs_i, (rhs_size, rhs_i), tree);
    node::typed_value_signed(lhs_size + rhs_size, val, mode, tree)
}

pub fn eval_shift(sources: &SourceSet,
                  tree: &mut Tree,
                  index: Index)
                  -> ResultIndex {
    let (mut kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_shift(tree, index));
    let lhs_tv = eval::deref_const_tv_check_z(tree, lhs_i)
        .or_else(|_| {
                     gen_err!(tree[lhs_i],
                              sources,
                              "shift operand cannot have z elements")
                 })?;
    let rhs_tv = eval::deref_const_tv_check_z(tree, rhs_i)
        .or_else(|_| {
                     gen_err!(tree[rhs_i],
                              sources,
                              "shift operand cannot have z elements")
                 })?;
    let (lhs_ty_i, lhs_val_i, rhs_ty_i, rhs_val_i) = match (lhs_tv, rhs_tv) {
        (Some((lt, lv)), Some((rt, rv))) => (lt, lv, rt, rv),
        _ => return shift_find_type(sources, tree, index, lhs_i, rhs_i),
    };
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    if !rhs_type.is_integral() {
        gen_err!(tree[rhs_i],
                 sources,
                 "can only shift by an integer type, not {}",
                 tree[rhs_ty_i].type_string())?;
    }
    let shift_bits = {
        let (rhs_val, _) = node::get_integer_val(tree, rhs_val_i).unwrap();
        let (sign, abs) = match eval::get_sign_abs(rhs_val,
                                                   rhs_type,
                                                   node::MAX_UNSIGNED_SIZE) {
            Some(sa) => sa,
            None => {
                gen_err!(tree[rhs_val_i],
                         sources,
                         "shift operand is too large")?
            }
        };
        if sign {
            kind = match kind {
                TokenKind::ShiftLeft => TokenKind::ShiftRight,
                TokenKind::ShiftRight => TokenKind::ShiftLeft,
                _ => unreachable!(),
            };
        }
        abs
    };
    let result = match lhs_type {
        Num::Int => shift_i(kind, lhs_val_i, shift_bits, tree),
        Num::Uns(size) => shift_u(kind, (size, lhs_val_i), shift_bits, tree),
        Num::Sig(size) => shift_s(kind, (size, lhs_val_i), shift_bits, tree),
        Num::Rat => shift_r(kind, lhs_val_i, shift_bits, tree),
        Num::Flo => shift_f(kind, lhs_val_i, shift_bits, tree),
        _ => {
            gen_err!(tree[lhs_val_i], sources, "can only shift numeric types")?
        }
    };
    Ok(Some(eval::replace_with(tree, index, result)))
}

fn shift_find_type(sources: &SourceSet,
                   tree: &mut Tree,
                   index: Index,
                   lhs_i: Index,
                   rhs_i: Index)
                   -> ResultIndex {
    let lhs_ty_i = eval::determine_type(sources, tree, lhs_i)?;
    let rhs_ty_i = eval::determine_type(sources, tree, rhs_i)?;
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    if !rhs_type.is_integral() {
        gen_err!(tree[rhs_i],
                 sources,
                 "can only shift by an integer type, not {}",
                 tree[rhs_ty_i].type_string())?;
    }
    match lhs_type {
        Num::Int | Num::Uns(_) | Num::Sig(_) | Num::Rat | Num::Flo => {
            Ok(Some(eval::set_type(tree, index, lhs_ty_i)))
        }
        _ => gen_err!(tree[lhs_i], sources, "can only shift numeric types"),
    }
}

fn shift_int(op: TokenKind,
             (val, mode): (Integer, Integer),
             bits: u32)
             -> (Integer, Integer) {
    match op {
        TokenKind::ShiftLeft => (val << bits, mode << bits),
        TokenKind::ShiftRight => (val >> bits, mode >> bits),
        _ => unreachable!(),
    }
}

fn shift_i(op: TokenKind, lhs_i: Index, bits: u32, tree: &mut Tree) -> Index {
    let val_mode = {
        let (val, mode) = node::get_integer_val(tree, lhs_i).unwrap();
        (val.clone(), mode.clone())
    };
    let (val, mode) = shift_int(op, val_mode, bits);
    node::typed_value_integer(val, mode, tree)
}

fn shift_u(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           bits: u32,
           tree: &mut Tree)
           -> Index {
    let val_mode = {
        let (val, mode) = node::get_integer_val(tree, lhs_i).unwrap();
        (val.clone(), mode.clone())
    };
    let (val, mode) = shift_int(op, val_mode, bits);
    node::typed_value_unsigned(lhs_size, val, mode, tree)
}

fn shift_s(op: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           bits: u32,
           tree: &mut Tree)
           -> Index {
    let val_mode = {
        let (val, mode) = node::get_integer_val(tree, lhs_i).unwrap();
        (eval::sign_extend(val, lhs_size), eval::sign_extend(mode, lhs_size))
    };
    let (val, mode) = shift_int(op, val_mode, bits);
    node::typed_value_signed(lhs_size, val, mode, tree)
}

fn shift_r(op: TokenKind, lhs_i: Index, bits: u32, tree: &mut Tree) -> Index {
    let val = node::get_rational_val(tree, lhs_i).unwrap().clone();
    let val = match op {
        TokenKind::ShiftLeft => val << bits,
        TokenKind::ShiftRight => val >> bits,
        _ => unreachable!(),
    };
    node::typed_value_rational(val, tree)
}

fn shift_f(op: TokenKind, lhs_i: Index, bits: u32, tree: &mut Tree) -> Index {
    let val = node::get_float_val(tree, lhs_i).unwrap().clone();
    let val = match op {
        TokenKind::ShiftLeft => val << bits,
        TokenKind::ShiftRight => val >> bits,
        _ => unreachable!(),
    };
    node::typed_value_float(val, tree)
}
