// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use node;
use tree::{Index, Link, Tree};

pub fn init(tree: &mut Tree, scope: Index) {
    Link::child(scope)
        .link(insert_integer_type(tree, scope), tree)
        .link(insert_bit_type(tree, scope), tree)
        .link(insert_unsigned_type_function(tree, scope), tree)
        .link(insert_signed_type_function(tree, scope), tree)
        .link(insert_rational_type(tree, scope), tree)
        .link(insert_float_type(tree, scope), tree)
        .link(insert_false_val(tree, scope), tree)
        .link(insert_true_val(tree, scope), tree);
}

pub type IntrinsicFunction = fn(tree: &mut Tree, inputs: Index)
                                -> Result<Index, Index>;

pub fn intrinsic_function(ident: &str) -> IntrinsicFunction {
    match ident {
        "make_unsigned_type" => make_unsigned_type,
        "make_signed_type" => make_signed_type,
        _ => unreachable!(),
    }
}

fn insert_integer_type(tree: &mut Tree, scope: Index) -> Index {
    let ty = node::type_type(tree);
    let val = node::integer_type(tree);
    let tv = node::typed_value(ty, val, tree);
    node::definition("integer".to_string(), scope, tv, tree)
}

fn insert_bit_type(tree: &mut Tree, scope: Index) -> Index {
    let ty = node::type_type(tree);
    let val = node::unsigned_type(1, tree);
    let tv = node::typed_value(ty, val, tree);
    node::definition("bit".to_string(), scope, tv, tree)
}

fn insert_unsigned_type_function(tree: &mut Tree, scope: Index) -> Index {
    insert_sized_type_function(tree, scope, false)
}

fn insert_signed_type_function(tree: &mut Tree, scope: Index) -> Index {
    insert_sized_type_function(tree, scope, true)
}

fn insert_sized_type_function(tree: &mut Tree,
                              scope: Index,
                              signed: bool)
                              -> Index {
    let (intrinsic_ident, def_ident) = if signed {
        ("make_signed_type".to_string(), "signed".to_string())
    } else {
        ("make_unsigned_type".to_string(), "unsigned".to_string())
    };
    let inputs = node::list(tree);
    let in_ident = "size".to_string();
    let in_ty = node::integer_type(tree);
    let in_field = node::field_def(in_ident, in_ty, tree);
    Link::child(inputs).link(in_field, tree);
    let outputs = node::type_type(tree);
    let port_scope = node::scope(Some(scope), tree);
    let stat_scope = node::scope(Some(port_scope), tree);
    let stat = node::stat_block(stat_scope, tree);
    let intrinsic = node::intrinsic(intrinsic_ident, tree);
    Link::sibling(stat_scope).link(intrinsic, tree);
    let func_ty = node::function_type(tree);
    let func = node::function(inputs, outputs, port_scope, stat, tree);
    let tv = node::typed_value(func_ty, func, tree);
    node::definition(def_ident, scope, tv, tree)
}

fn insert_rational_type(tree: &mut Tree, scope: Index) -> Index {
    let ty = node::type_type(tree);
    let val = node::rational_type(tree);
    let tv = node::typed_value(ty, val, tree);
    node::definition("rational".to_string(), scope, tv, tree)
}

fn insert_float_type(tree: &mut Tree, scope: Index) -> Index {
    let ty = node::type_type(tree);
    let val = node::float_type(tree);
    let tv = node::typed_value(ty, val, tree);
    node::definition("float".to_string(), scope, tv, tree)
}

fn insert_false_val(tree: &mut Tree, scope: Index) -> Index {
    let tv = node::typed_value_bit(false, tree);
    node::definition("false".to_string(), scope, tv, tree)
}

fn insert_true_val(tree: &mut Tree, scope: Index) -> Index {
    let tv = node::typed_value_bit(true, tree);
    node::definition("true".to_string(), scope, tv, tree)
}

fn make_unsigned_type(tree: &mut Tree, inputs: Index) -> Result<Index, Index> {
    make_sized_type(tree, inputs, false)
}

fn make_signed_type(tree: &mut Tree, inputs: Index) -> Result<Index, Index> {
    make_sized_type(tree, inputs, true)
}

fn make_sized_type(tree: &mut Tree,
                   inputs: Index,
                   signed: bool)
                   -> Result<Index, Index> {
    let input = node::get_list(tree, inputs)
        .unwrap()
        .next(tree)
        .unwrap();
    let size = {
        let (val, mode) = node::get_integer_val(tree, input).unwrap();
        let label = if signed { "signed" } else { "unsigned" };
        if *mode != 0 {
            Err(format!("size of {}() cannot be x or z", label))
        } else if *val < 0 {
            Err(format!("size of {}() cannot be negative", label))
        } else if *val > node::MAX_UNSIGNED_SIZE {
            Err(format!("size of {}() cannot be larger than {}",
                        label,
                        node::MAX_UNSIGNED_SIZE))
        } else {
            Ok(val.to_u32_wrapping())
        }
    };
    let size = match size {
        Ok(size) => size,
        Err(message) => {
            let err = node::err_message(message, tree);
            return Err(err);
        }
    };
    let ret = if signed {
        node::signed_type(size, tree)
    } else {
        node::unsigned_type(size, tree)
    };
    Ok(ret)
}
