// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#[macro_use]
extern crate clap;
extern crate rugint;
extern crate rugrat;
extern crate rugflo;

use clap::Arg;
use std::error::Error;
use std::fs;
use std::io::{self, Write};
use std::process;

macro_rules! unless_ret_none {
    ($e:expr) => {
        if !$e {
            return None;
        }
    }
}

macro_rules! unwrap_or_ret {
    ($e:expr) => {
        unwrap_or_ret!($e, ())
    };

    ($e:expr, $r:expr) => {{
        let mut value = None;
        $e.map(|x| value = Some(x));
        match value {
            Some(x) => x,
            None => return $r,
        }
    }};
}

macro_rules! gen_err {
    ($sn:expr, $s:expr, $m:expr) => {
        Err(if $sn.span().check().is_none() {
            Error::general(&$m)
        } else {
            Error::general(&format!("{}: {}", $sn.span().loc_string($s), $m))
        })
    };
    ($sn:expr, $s:expr, $f:expr $(, $p:expr)+) => {
        gen_err!($sn, $s, &format!($f $(, $p)+))
    };
}

#[macro_use]
mod try_next;

mod arith;
mod bitwise;
mod eval;
mod core;
mod error;
mod idents;
mod input;
mod rel;
mod node;
mod output;
mod package;
mod parse;
mod tree;

struct Options<'a> {
    filenames: &'a [&'a str],
    dump_parse: Option<&'a str>,
    dump_resolve: Option<&'a str>,
    dump_eval: Option<&'a str>,
    dump_error: Option<&'a str>,
}

fn err_out(filename: &str, err: io::Error) -> error::Error {
    let filename = if filename == "" {
        "<standard output>"
    } else {
        filename
    };
    error::Error::io(filename, err)
}

fn get_writer(filename: &str) -> Result<Box<Write>, error::Error> {
    if filename == "" {
        Ok(Box::new(std::io::stdout()))
    } else {
        match fs::File::create(filename).map(io::BufWriter::new) {
            Ok(writer) => Ok(Box::new(writer)),
            Err(e) => Err(err_out(filename, e)),
        }
    }
}

fn optional_dump(dump: Option<&str>,
                 package: &package::Package)
                 -> Result<(), error::Error> {
    if let Some(filename) = dump {
        let mut writer = get_writer(filename)?;
        write!(writer, "{:?}", package)
            .map_err(|e| err_out(filename, e))?;
    }
    Ok(())
}

fn translate(options: Options) -> Result<(), error::Error> {
    let mut package = package::Package::new();
    let ret = package
        .parse_sources(options.filenames)
        .and_then(|_| optional_dump(options.dump_parse, &package))
        .and_then(|_| package.resolve_idents())
        .and_then(|_| optional_dump(options.dump_resolve, &package))
        .and_then(|_| package.evaluate())
        .and_then(|_| optional_dump(options.dump_eval, &package))
        .and_then(|_| package.output());
    if ret.is_err() {
        let _ = optional_dump(options.dump_error, &package);
    }
    ret
}

fn main() {
    let app = app_from_crate!("\n")
        .arg(Arg::with_name("INPUT")
                 .help("The input files to translate")
                 .required(true)
                 .multiple(true)
                 .index(1))
        .arg(Arg::with_name("dump-parse")
                 .long("dump-parse")
                 .takes_value(true)
                 .value_names(&["OUTPUT FILENAME"])
                 .help("Dumps the tree after parsing"))
        .arg(Arg::with_name("dump-resolve")
                 .long("dump-resolve")
                 .takes_value(true)
                 .value_names(&["OUTPUT FILENAME"])
                 .help("Dumps the tree after resolving identifiers"))
        .arg(Arg::with_name("dump-evaluate")
                 .long("dump-evaluate")
                 .takes_value(true)
                 .value_names(&["OUTPUT FILENAME"])
                 .help("Dumps the tree after evaluating expressions"))
        .arg(Arg::with_name("dump-error")
                 .long("dump-error")
                 .takes_value(true)
                 .value_names(&["OUTPUT FILENAME"])
                 .help("Dumps the tree if an error occurs"));
    let matches = app.get_matches();
    let options = Options {
        filenames: &matches.values_of("INPUT").unwrap().collect::<Vec<_>>(),
        dump_parse: matches.value_of("dump-parse"),
        dump_resolve: matches.value_of("dump-resolve"),
        dump_eval: matches.value_of("dump-evaluate"),
        dump_error: matches.value_of("dump-error"),
    };
    match translate(options) {
        Ok(_) => {}
        Err(e) => {
            writeln!(&mut io::stderr(), "{}", e.description()).unwrap();
            process::exit(2);
        }
    }
}
