// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use error::Error;
use eval::{self, Num, ResultIndex};
use input::SourceSet;
use node;
use parse::TokenKind;
use rugint::{Assign, Integer};
use std::u32;
use tree::{Index, Link, Tree};

pub fn eval(sources: &SourceSet,
            tree: &mut Tree,
            index: Index,
            scope_i: Index)
            -> ResultIndex {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_relational(tree, index));
    let lhs_tv = eval::deref_const_tv_check_z(tree, lhs_i)
        .or_else(|_| {
                     gen_err!(tree[lhs_i],
                              sources,
                              "relational operand cannot have z elements")
                 })?;
    let rhs_tv = eval::deref_const_tv_check_z(tree, rhs_i)
        .or_else(|_| {
                     gen_err!(tree[rhs_i],
                              sources,
                              "relational operand cannot have z elements")
                 })?;
    let (lhs_ty_i, lhs_val_i, rhs_ty_i, rhs_val_i) = match (lhs_tv, rhs_tv) {
        (Some((lt, lv)), Some((rt, rv))) => (lt, lv, rt, rv),
        _ => {
            return rel_find_type(sources,
                                 tree,
                                 index,
                                 scope_i,
                                 (kind, lhs_i, rhs_i))
        }
    };
    let lhs_type = match eval::num_type(tree, lhs_ty_i) {
        Num::Uns(_) => Num::Int,
        n => n,
    };
    let rhs_type = match eval::num_type(tree, rhs_ty_i) {
        Num::Uns(_) => Num::Int,
        n => n,
    };
    let result = match (lhs_type, rhs_type) {
        (Num::Int, Num::Int) => cmp_i_i(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Int, Num::Sig(r)) => {
            cmp_i_s(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Int, Num::Rat) => cmp_i_r(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Sig(l), Num::Int) => {
            cmp_i_s(kind.reverse_rel(), rhs_val_i, (l, lhs_val_i), tree)
        }
        (Num::Sig(l), Num::Sig(r)) => {
            cmp_s_s(kind, (l, lhs_val_i), (r, rhs_val_i), tree)
        }
        (Num::Sig(l), Num::Rat) => {
            cmp_r_s(kind.reverse_rel(), rhs_val_i, (l, lhs_val_i), tree)
        }
        (Num::Rat, Num::Int) => {
            cmp_i_r(kind.reverse_rel(), rhs_val_i, lhs_val_i, tree)
        }
        (Num::Rat, Num::Sig(r)) => {
            cmp_r_s(kind, lhs_val_i, (r, rhs_val_i), tree)
        }
        (Num::Rat, Num::Rat) => cmp_r_r(kind, lhs_val_i, rhs_val_i, tree),
        (Num::Flo, Num::Flo) => cmp_f_f(kind, lhs_val_i, rhs_val_i, tree),
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot compare values of types {} and {}",
                     tree[lhs_ty_i].type_string(),
                     tree[rhs_ty_i].type_string())?
        }
    };
    let val_i = node::typed_value_bit(result, tree);
    Ok(Some(eval::replace_with(tree, index, val_i)))
}

#[derive(PartialEq)]
struct IsEmpty(bool);

fn get_ident_def(tree: &mut Tree,
                 scope_i: Index,
                 var_i: Index)
                 -> (Index, Index) {
    if let Some(def_i) = node::get_resolved_ident(tree, var_i) {
        (var_i, def_i)
    } else {
        let name = "$rel_slice".to_string();
        let def_i = node::definition(name, scope_i, var_i, tree);
        Link::child(scope_i).link(def_i, tree);
        let ident_i = node::resolved_ident(def_i, tree);
        (ident_i, def_i)
    }
}

fn rel_check_x(tree: &mut Tree,
               scope_i: Index,
               lhs_i: Index,
               rhs_i: Index,
               num_ty: Num)
               -> IsEmpty {
    let lhs_tv = eval::deref_const_tv(tree, lhs_i);
    let rhs_tv = eval::deref_const_tv(tree, rhs_i);
    let (const_ty_i, const_val_i, var_i) = match (lhs_tv, rhs_tv) {
        (None, None) => return IsEmpty(false),
        (Some((l_ty_i, l_val_i)), None) => (l_ty_i, l_val_i, rhs_i),
        (None, Some((r_ty_i, r_val_i))) => (r_ty_i, r_val_i, lhs_i),
        (Some(_), Some(_)) => unreachable!(),
    };
    let (size, signed) = match num_ty {
        Num::Uns(size) => (size, false),
        Num::Sig(size) => (size, true),
        _ => unreachable!(),
    };
    // we know there are no z bits, so just check for x bits
    let (mut new_const_val, x_mask): (_, Integer) = {
        let (val, mode) = node::get_integer_val(tree, const_val_i).unwrap();
        (!mode.clone() & val, (eval::shl_one(size) - 1) & mode)
    };
    let mut new_var_i = None;
    let mut orig_def = None;
    let mut start;
    let mut end = 0;
    let mut curr_size = 0;
    let var_link = tree.replacement_link(var_i);
    let mut lower_part = Integer::from(0);
    loop {
        start = match x_mask.find_zero(end.into()) {
            Some(s) if s < size => s,
            _ => break,
        };
        lower_part.assign(1);
        lower_part = ((lower_part << curr_size) - 1) & &new_const_val;
        new_const_val = ((new_const_val >> (start - end + curr_size)) <<
                         curr_size) | &lower_part;
        end = match x_mask.find_one(start.into()) {
            Some(s) if s < size => s,
            _ => size,
        };
        let ident_i = match orig_def {
            None => {
                let (ident_i, def_i) = get_ident_def(tree, scope_i, var_i);
                orig_def = Some(def_i);
                ident_i
            }
            Some(def_i) => node::resolved_ident(def_i, tree),
        };
        let range_i = node::range_decoded(start, end, tree);
        let expr_i = node::expr_index(ident_i, range_i, tree);
        let ty_i = if signed {
            node::signed_type(end - start, tree)
        } else {
            node::unsigned_type(end - start, tree)
        };
        let tv_i = eval::set_type(tree, expr_i, ty_i);
        curr_size += end - start;
        new_var_i = match new_var_i {
            None => Some(tv_i),
            Some(current_i) => {
                let expr_i = node::expr_concatenation(tv_i, current_i, tree);
                let ty_i = if signed {
                    node::signed_type(curr_size, tree)
                } else {
                    node::unsigned_type(curr_size, tree)
                };
                Some(eval::set_type(tree, expr_i, ty_i))
            }
        };
    }
    match new_var_i {
        Some(var_i) => {
            var_link.unwrap().link(var_i, tree);
        }
        _ => return IsEmpty(true),
    }
    let new_const_val_i = node::integer_val(new_const_val, 0.into(), tree);
    eval::replace_with(tree, const_val_i, new_const_val_i);
    let new_const_ty_i = if signed {
        node::signed_type(curr_size, tree)
    } else {
        node::unsigned_type(curr_size, tree)
    };
    eval::replace_with(tree, const_ty_i, new_const_ty_i);
    IsEmpty(false)
}

fn rel_find_type(sources: &SourceSet,
                 tree: &mut Tree,
                 index: Index,
                 scope_i: Index,
                 (kind, lhs_i, rhs_i): (TokenKind, Index, Index))
                 -> ResultIndex {
    let lhs_ty_i = eval::determine_type(sources, tree, lhs_i)?;
    let rhs_ty_i = eval::determine_type(sources, tree, rhs_i)?;
    let lhs_type = eval::num_type(tree, lhs_ty_i);
    let rhs_type = eval::num_type(tree, rhs_ty_i);
    let is_empty = match (lhs_type, rhs_type) {
        (Num::Int, Num::Int) |
        (Num::Int, Num::Rat) |
        (Num::Rat, Num::Int) |
        (Num::Rat, Num::Rat) |
        (Num::Flo, Num::Flo) => IsEmpty(false),
        (Num::Uns(l), Num::Uns(r)) |
        (Num::Sig(l), Num::Sig(r)) => {
            let (new_ty_i, new_lhs_i, new_rhs_i) =
                eval::cast_to_bigger_ty_tops(sources,
                                             tree,
                                             (l, lhs_i, lhs_ty_i),
                                             (r, rhs_i, rhs_ty_i))?;
            let new_type = eval::num_type(tree, new_ty_i);
            rel_check_x(tree, scope_i, new_lhs_i, new_rhs_i, new_type)
        }
        (Num::Uns(u), Num::Sig(s)) |
        (Num::Sig(s), Num::Uns(u)) if u >= s => {
            let ty_i = node::signed_type(u + 1, tree);
            let new_type = eval::num_type(tree, ty_i);
            let new_lhs_i = eval::cast_type(sources, tree, lhs_i, ty_i)?;
            let new_rhs_i = eval::cast_type(sources, tree, rhs_i, ty_i)?;
            rel_check_x(tree, scope_i, new_lhs_i, new_rhs_i, new_type)
        }
        (Num::Int, Num::Uns(_)) |
        (Num::Int, Num::Sig(_)) |
        (Num::Uns(_), Num::Sig(_)) => {
            let new_lhs_i = eval::cast_type(sources, tree, lhs_i, rhs_ty_i)?;
            rel_check_x(tree, scope_i, new_lhs_i, rhs_i, rhs_type)
        }
        (Num::Uns(_), Num::Int) |
        (Num::Sig(_), Num::Int) |
        (Num::Sig(_), Num::Uns(_)) => {
            let new_rhs_i = eval::cast_type(sources, tree, rhs_i, lhs_ty_i)?;
            rel_check_x(tree, scope_i, lhs_i, new_rhs_i, lhs_type)
        }
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cannot compare values of types {} and {}",
                     tree[lhs_ty_i].type_string(),
                     tree[rhs_ty_i].type_string())?
        }
    };
    match is_empty {
        IsEmpty(true) => {
            let result = cmp_gen(kind, &0, &0);
            let node_i = node::typed_value_bit(result, tree);
            Ok(Some(eval::replace_with(tree, index, node_i)))
        }
        IsEmpty(false) => {
            let ty_i = node::bit_type(tree);
            Ok(Some(eval::set_type(tree, index, ty_i)))
        }
    }
}

// lhs  rhs  operation
// v m  v m  lhs op rhs
//   0    0  l_v op r_v
//   0  0 1  l_v op l_v
// 0 1    0  r_v op r_v
// 0 1  0 1  0   op 0
// 1 1       unreachable
//      1 1  unreachable
// So, operation lhs = l_v or (l_m and r_v)
//     operation rhs = r_v or (r_m and l_v)

fn cmp_gen<Lhs, Rhs>(rel: TokenKind, lhs: &Lhs, rhs: &Rhs) -> bool
    where Lhs: PartialEq<Rhs>,
          Lhs: PartialOrd<Rhs>
{
    match rel {
        TokenKind::IsEquals => lhs == rhs,
        TokenKind::NotEquals => lhs != rhs,
        TokenKind::Less => lhs < rhs,
        TokenKind::LessEquals => lhs <= rhs,
        TokenKind::Greater => lhs > rhs,
        TokenKind::GreaterEquals => lhs >= rhs,
        _ => unreachable!(),
    }
}

fn cmp_i_i(rel: TokenKind, lhs_i: Index, rhs_i: Index, tree: &Tree) -> bool {
    let (lhs_v, lhs_m) = node::get_integer_val(tree, lhs_i).unwrap();
    let (rhs_v, rhs_m) = node::get_integer_val(tree, rhs_i).unwrap();
    let lhs = (lhs_m.clone() & rhs_v) | lhs_v;
    let rhs = (rhs_m.clone() & lhs_v) | rhs_v;
    cmp_gen(rel, &lhs, &rhs)
}

fn cmp_i_s(rel: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &Tree)
           -> bool {
    let (lhs_v, lhs_m) = node::get_integer_val(tree, lhs_i).unwrap();
    let (rhs_v, rhs_m) = node::get_integer_val(tree, rhs_i).unwrap();
    let rhs_v_ext = eval::sign_extend(rhs_v, rhs_size);
    let rhs_m_ext = eval::sign_extend(rhs_m, rhs_size);
    let rhs = (rhs_m_ext & lhs_v) | &rhs_v_ext;
    let lhs = (rhs_v_ext & lhs_m) | lhs_v;
    cmp_gen(rel, &lhs, &rhs)
}

fn cmp_i_r(rel: TokenKind, lhs_i: Index, rhs_i: Index, tree: &Tree) -> bool {
    let (lhs, _) = node::get_integer_val(tree, lhs_i).unwrap();
    let rhs = node::get_rational_val(tree, rhs_i).unwrap();
    cmp_gen(rel, lhs, rhs)
}

fn cmp_s_s(rel: TokenKind,
           (lhs_size, lhs_i): (u32, Index),
           (rhs_size, rhs_i): (u32, Index),
           tree: &Tree)
           -> bool {
    let (lhs_v_ext, lhs_m_ext);
    let (mut lhs_v, mut lhs_m) = node::get_integer_val(tree, lhs_i).unwrap();
    lhs_v_ext = eval::sign_extend_if_neg(lhs_v, lhs_size);
    if let Some(ref l_v) = lhs_v_ext {
        lhs_v = l_v;
    }
    lhs_m_ext = eval::sign_extend_if_neg(lhs_m, lhs_size);
    if let Some(ref l_m) = lhs_m_ext {
        lhs_m = l_m;
    }
    let (rhs_v, rhs_m) = node::get_integer_val(tree, rhs_i).unwrap();
    let rhs_v_ext = eval::sign_extend(rhs_v, rhs_size);
    let rhs_m_ext = eval::sign_extend(rhs_m, rhs_size);
    let rhs = (rhs_m_ext & lhs_v) | &rhs_v_ext;
    let lhs = (rhs_v_ext & lhs_m) | lhs_v;
    cmp_gen(rel, &lhs, &rhs)
}

fn cmp_r_s(rel: TokenKind,
           lhs_i: Index,
           (rhs_size, rhs_i): (u32, Index),
           tree: &Tree)
           -> bool {
    let lhs = node::get_rational_val(tree, lhs_i).unwrap();
    let extended_rhs;
    let (mut rhs, _) = node::get_integer_val(tree, rhs_i).unwrap();
    if let Some(neg) = eval::sign_extend_if_neg(rhs, rhs_size) {
        extended_rhs = neg;
        rhs = &extended_rhs
    }
    cmp_gen(rel, lhs, rhs)
}

fn cmp_r_r(rel: TokenKind, lhs_i: Index, rhs_i: Index, tree: &Tree) -> bool {
    let lhs = node::get_rational_val(tree, lhs_i).unwrap();
    let rhs = node::get_rational_val(tree, rhs_i).unwrap();
    cmp_gen(rel, lhs, rhs)
}

fn cmp_f_f(rel: TokenKind, lhs_i: Index, rhs_i: Index, tree: &Tree) -> bool {
    let lhs = node::get_float_val(tree, lhs_i).unwrap();
    let rhs = node::get_float_val(tree, rhs_i).unwrap();
    cmp_gen(rel, lhs, rhs)
}
