// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use input::SourceSet;
use node::Node;
use std::fmt;
use std::ops;

#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq, PartialOrd)]
// must never be zero
pub struct Index {
    val: u32,
}

impl Index {
    pub fn as_u32(self) -> u32 {
        self.val
    }
}

#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq, PartialOrd)]
pub struct OptionIndex {
    optional: u32,
}

impl OptionIndex {
    pub fn as_option(&self) -> Option<Index> {
        match self.optional {
            0 => None,
            val => Some(Index { val: val }),
        }
    }
}

impl PartialEq<Index> for OptionIndex {
    fn eq(&self, other: &Index) -> bool {
        self.optional == other.val
    }
}

impl PartialEq<OptionIndex> for Index {
    fn eq(&self, other: &OptionIndex) -> bool {
        self.val == other.optional
    }
}

impl From<Index> for OptionIndex {
    fn from(index: Index) -> OptionIndex {
        OptionIndex { optional: index.val }
    }
}

impl From<Option<Index>> for OptionIndex {
    fn from(option: Option<Index>) -> OptionIndex {
        match option {
            Some(index) => index.into(),
            None => OptionIndex { optional: 0 },
        }
    }
}

impl fmt::Display for Index {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        self.val.fmt(f)
    }
}

struct TreeNode {
    node: Node,
    back_ref: OptionIndex,
    child: OptionIndex,
    sibling: OptionIndex,
}

#[derive(Clone, Copy)]
pub struct Siblings {
    index: OptionIndex,
}

impl Default for Siblings {
    fn default() -> Siblings {
        Siblings { index: None.into() }
    }
}

// Siblings has to store the next element internally, so that this
// works:
// ```rust
// let r = siblings.next(tree).unwrap();
// tree.remove(r);
// let n = siblings.next(tree).unwrap();
// ```
impl Siblings {
    pub fn iter<'a>(&self, tree: &'a Tree) -> SiblingsIterator<'a> {
        SiblingsIterator {
            siblings: *self,
            tree: self.index.as_option().map(|_| tree),
        }
    }

    pub fn next(&mut self, tree: &Tree) -> Option<Index> {
        let oi = self.index.as_option();
        if let Some(i) = oi {
            self.index = tree.next_sibling(i).into();
        }
        oi
    }

    pub fn peek(&mut self) -> Option<Index> {
        self.index.as_option()
    }
}

pub struct SiblingsIterator<'a> {
    siblings: Siblings,
    tree: Option<&'a Tree>,
}

impl<'a> Iterator for SiblingsIterator<'a> {
    type Item = Index;

    fn next(&mut self) -> Option<Index> {
        if let Some(tree) = self.tree {
            let ret = self.siblings.next(tree);
            if ret.is_none() {
                self.tree = None;
            }
            ret
        } else {
            None
        }
    }
}

// Internal:
//     tree[0] = {
//         node: default
//         back_ref: None
//         child: None
//         sibling: None-terminated linked list of free nodes to reuse
//     }
//     tree[<free node>] = {
//         node: default
//         back_ref: self index
//         child: None
//         sibling: None-terminated linked list of free nodes to reuse
//     }
pub struct Tree(Vec<TreeNode>);

#[derive(Copy, Clone)]
enum LinkRelation {
    Child,
    Sibling,
}

#[derive(Copy, Clone)]
pub struct Link {
    relation: LinkRelation,
    index: Index,
}

impl Link {
    pub fn child(index: Index) -> Link {
        Link {
            relation: LinkRelation::Child,
            index: index,
        }
    }

    pub fn sibling(index: Index) -> Link {
        Link {
            relation: LinkRelation::Sibling,
            index: index,
        }
    }

    pub fn link(&mut self, index: Index, tree: &mut Tree) -> &mut Link {
        tree.make_root(index);
        match self.relation {
            LinkRelation::Child => {
                let parent = self.index;
                tree.assert_live(parent);
                assert!(parent != index, "node cannot become its own child");
                let child = tree.i(parent).child;
                tree.i_mut(index).back_ref = parent.into();
                tree.i_mut(parent).child = index.into();
                tree.i_mut(index).sibling = child;
                if let Some(child) = child.as_option() {
                    tree.i_mut(child).back_ref = index.into();
                }
            }
            LinkRelation::Sibling => {
                let previous_sibling = self.index;
                tree.assert_live(previous_sibling);
                assert!(previous_sibling != index,
                        "node cannot become its own sibling");
                let next_sibling = tree.i(previous_sibling).sibling;
                tree.i_mut(index).back_ref = previous_sibling.into();
                tree.i_mut(previous_sibling).sibling = index.into();
                tree.i_mut(index).sibling = next_sibling;
                if let Some(next_sibling) = next_sibling.as_option() {
                    tree.i_mut(next_sibling).back_ref = index.into();
                }
            }
        }
        *self = Link::sibling(index);
        self
    }

    pub fn get_sibling(&self) -> Option<Index> {
        match self.relation {
            LinkRelation::Sibling => Some(self.index),
            _ => None,
        }
    }
}

impl Tree {
    pub fn new() -> Tree {
        let mut buf = Vec::with_capacity(64);
        buf.push(TreeNode {
                     node: Default::default(),
                     back_ref: None.into(),
                     child: None.into(),
                     sibling: None.into(),
                 });
        Tree(buf)
    }

    pub fn next_insert_index(&self) -> Index {
        self.free()
            .as_option()
            .unwrap_or_else(|| Index { val: self.0.len() as u32 })
    }

    pub fn insert(&mut self, node: Node) -> Index {
        let tree_node = TreeNode {
            node: node,
            back_ref: None.into(),
            child: None.into(),
            sibling: None.into(),
        };
        if let Some(free_index) = self.free().as_option() {
            let next_free = self.i(free_index).sibling;
            self.set_free(next_free);
            *self.i_mut(free_index) = tree_node;
            free_index
        } else {
            let index = Index { val: self.0.len() as u32 };
            self.0.push(tree_node);
            index
        }
    }

    pub fn remove(&mut self, index: Index) {
        self.make_root(index);
        self.free_node(index);
    }

    pub fn replacement_link(&self, index: Index) -> Option<Link> {
        self.assert_live(index);
        if let Some(back_ref) = self.i(index).back_ref.as_option() {
            let relation = if self.i(back_ref).child == index {
                LinkRelation::Child
            } else {
                LinkRelation::Sibling
            };
            Some(Link {
                     relation: relation,
                     index: back_ref,
                 })
        } else {
            None
        }
    }

    pub fn is_root(&self, index: Index) -> bool {
        self.assert_live(index);
        self.i(index).back_ref.as_option().is_none()
    }

    pub fn make_root(&mut self, index: Index) {
        self.assert_live(index);
        let back_ref = self.i(index).back_ref;
        let sibling = self.i(index).sibling;
        if let Some(back_ref) = back_ref.as_option() {
            if self.i(back_ref).child == index {
                self.i_mut(back_ref).child = sibling
            } else {
                self.i_mut(back_ref).sibling = sibling;
            }
        }
        if let Some(sibling) = sibling.as_option() {
            self.i_mut(sibling).back_ref = back_ref;
        }
        self.i_mut(index).back_ref = None.into();
        self.i_mut(index).sibling = None.into();
    }

    fn free_node(&mut self, index: Index) {
        let mut child = self.i(index).child;
        while let Some(child_index) = child.as_option() {
            let next_child = self.i(child_index).sibling;
            self.free_node(child_index);
            child = next_child;
        }
        self.i_mut(index).node = Default::default();
        self.i_mut(index).back_ref = index.into();
        self.i_mut(index).child = None.into();
        self.i_mut(index).sibling = self.free();
        self.set_free(index.into());
    }

    pub fn parent(&self, mut child: Index) -> Option<Index> {
        self.assert_live(child);
        loop {
            let back_ref = match self.i(child).back_ref.as_option() {
                Some(back_ref) => back_ref,
                None => return None,
            };
            if self.i(back_ref).child == child {
                return Some(back_ref);
            }
            child = back_ref;
        }
    }

    pub fn first_child(&self, parent: Index) -> Option<Index> {
        self.assert_live(parent);
        self.i(parent).child.as_option()
    }

    pub fn children(&self, parent: Index) -> Siblings {
        self.assert_live(parent);
        Siblings { index: self.i(parent).child }
    }

    pub fn next_sibling(&self, sibling: Index) -> Option<Index> {
        self.assert_live(sibling);
        self.i(sibling).sibling.as_option()
    }

    pub fn next_siblings(&self, previous_sibling: Index) -> Siblings {
        self.assert_live(previous_sibling);
        Siblings { index: self.i(previous_sibling).sibling }
    }

    fn format(&self,
              fmt: &mut fmt::Formatter,
              sources: Option<&SourceSet>,
              index: Index,
              prefix: &str)
              -> fmt::Result {
        self.format_rec(fmt, sources, index, &mut prefix.to_string())
    }

    fn format_rec(&self,
                  fmt: &mut fmt::Formatter,
                  sources: Option<&SourceSet>,
                  index: Index,
                  prefix: &mut String)
                  -> fmt::Result {
        let sibling = self.next_sibling(index);
        let original_prefix_len = prefix.len();
        if sibling.is_some() {
            prefix.push_str("├── ")
        } else {
            prefix.push_str("└── ")
        };
        write!(fmt, "{}{}: ", prefix, index.val)?;
        self[index].format(fmt, sources)?;
        write!(fmt, "\n")?;
        prefix.truncate(original_prefix_len);
        if sibling.is_some() {
            prefix.push_str("│   ")
        } else {
            prefix.push_str("    ")
        };
        for child_i in self.children(index).iter(self) {
            self.format_rec(fmt, sources, child_i, prefix)?;
        }
        prefix.truncate(original_prefix_len);
        Ok(())
    }

    fn assert_live(&self, index: Index) {
        assert!(self.i(index).back_ref != index, "index out of bounds");
    }

    fn i(&self, index: Index) -> &TreeNode {
        &self.0[index.val as usize]
    }

    fn i_mut(&mut self, index: Index) -> &mut TreeNode {
        &mut self.0[index.val as usize]
    }

    fn free(&self) -> OptionIndex {
        self.0[0].sibling
    }

    fn set_free(&mut self, free: OptionIndex) {
        self.0[0].sibling = free;
    }

    pub fn display<'a>(&'a self,
                       sources: Option<&'a SourceSet>,
                       index: Index)
                       -> TreeDisplay {
        TreeDisplay {
            tree: self,
            sources: sources,
            index: index,
        }
    }
}

pub struct TreeDisplay<'a> {
    tree: &'a Tree,
    sources: Option<&'a SourceSet>,
    index: Index,
}

impl<'a> fmt::Display for TreeDisplay<'a> {
    fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
        self.tree.format(fmt, self.sources, self.index, "")
    }
}

impl ops::Index<Index> for Tree {
    type Output = Node;

    fn index(&self, index: Index) -> &Node {
        self.assert_live(index);
        &self.i(index).node
    }
}

impl ops::IndexMut<Index> for Tree {
    fn index_mut(&mut self, index: Index) -> &mut Node {
        self.assert_live(index);
        &mut self.i_mut(index).node
    }
}
